package denim;

import beast.core.*;
import beast.core.parameter.RealParameter;
import beast.evolution.alignment.TaxonSet;
import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;
import beast.evolution.tree.TreeDistribution;
import beast.evolution.tree.TreeInterface;
import denim.util.*;

import java.util.*;

/**
 *  Created by Graham Jones on 27/10/2016.
 */

public class CoalescenceMigrationDistribution extends TreeDistribution {
    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<GtreeWithEmbedding>> gTreeWEsInput =
            new Input<List<GtreeWithEmbedding>>("geneTree",
                    "All gene trees, with embedding parameters, and ploidy values.",
                    new ArrayList<GtreeWithEmbedding>());

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<InverseGammaComponent>> priorComponentsInput =
            new Input<List<InverseGammaComponent>>("popPriorInvGamma",
                    "Component of mixture of inverse gamma distributions used as a prior for the per-branch populations",
                    new ArrayList<InverseGammaComponent>(),
                    Input.Validate.REQUIRED);

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<RealParameter> popPriorScaleInput =
            new Input<RealParameter>("popPriorScale",
                    "Overall scale for population size",
                    Input.Validate.REQUIRED);

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<RealParameter> migrationIntensityInput =
            new Input<>("migrationIntensity",
                    "Migration intensity: A value v which determines how likely migrations are. The expected number" +
                            "of migrations in a gene tree with n tips is v(n-1), or v per pair of branches.",
                    Input.Validate.REQUIRED);

    // Needed for Beauti template, but not used here
    @SuppressWarnings("UnusedDeclaration")
    public Input<TaxonSet> taxonSetInput =
            new Input<TaxonSet>("taxonset",
                    "set of taxa mapping lineages to species", Input.Validate.REQUIRED);


    private Bindings bindings;

    private TreeInterface sTree;
    private int nSMCTreeNodes;
    private int nSMCTreeTips;
    private List<GtreeWithEmbedding> gTreesWE;
    private ArrayList<Tree> gTrees;
    private int nGTrees;

    private double [][] lnGammaRatiosTable;

    private CoalescencesMigrations coalsMigs;

    // This is the data that is stored and restored.

    private int [][] coalCounts;
    // coalCounts[b][j] = number of coalescences in branch b in gtree j.

    private int [][] migEventsCounts;
    // migEventsCounts[b][j] is the change in number of lineages due to migrations
    // in branch b in gtree j.

    private int [][] migLeaveCounts;
    // migEventsCounts[b][j] is the number of lineages leaving branch b in gtree j
    // via migration.

    // coalIntensities[b][j]  =  gamma_bj  =
    //      sum_i c_jbi (n_jbi choose 2)(1/p_j)
    // where c_jbi is i'th time interval and n_jbi are numbers of lineages in the interval,
    // and p_j is ploidy.
    private double [][] coalIntensities;



    // These are the stored version of the above.

    /*
    private int [][] stored_coalCounts;
    private int [][] stored_migEventsCounts;
    private int [][] stored_migLeaveCounts;
    private double [][] stored_coalIntensities;
*/

    private boolean debugFlag = Boolean.valueOf(System.getProperty("denim.debug"));
    private int numberofdebugchecks = 0;
    private final static int maxnumberofdebugchecks = 1000000;



    @Override
    public void initAndValidate() {
        super.initAndValidate(); // ??

        gTreesWE = gTreeWEsInput.get();
        nGTrees = gTreesWE.size();

        gTrees = new ArrayList<>(nGTrees);
        for (int j = 0; j < nGTrees; j++) {
            gTrees.add(gTreesWE.get(j).getTree());
        }

        sTree = treeInput.get();
        bindings = Bindings.initialise(sTree, gTrees);
        /*introgsCoals = new IntrogressionsCoalescences(sTree, gTreesWI, bindings);*/
        nSMCTreeNodes = sTree.getNodeCount();
        nSMCTreeTips = sTree.getLeafNodeCount();

        int totalGTipCount = 0;
        for (Tree gtree : gTrees) {
            totalGTipCount += gtree.getLeafNodeCount();
        }
        // number of coalescences and migrations is <= 2*totalGTipCount
        lnGammaRatiosTable = new double[priorComponentsInput.get().size()][2 * totalGTipCount];
        for (int c = 0; c < priorComponentsInput.get().size(); c++) {
            InverseGammaComponent igc = priorComponentsInput.get().get(c);
            for (int q = 0; q < lnGammaRatiosTable[c].length; q++) {
                lnGammaRatiosTable[c][q] = lnRatioGammas(igc.getAlpha(), q);
            }
        }
        coalsMigs = new CoalescencesMigrations(sTree, gTreesWE, bindings);

        coalCounts = new int[nSMCTreeNodes][nGTrees];
        migEventsCounts = new int[nSMCTreeNodes][nGTrees];
        migLeaveCounts = new int[nSMCTreeNodes][nGTrees];
        coalIntensities = new double[nSMCTreeNodes][nGTrees];

        /*
        stored_coalCounts = new int[nSMCTreeNodes][nGTrees];
        stored_migEventsCounts = new int[nSMCTreeNodes][nGTrees];
        stored_migLeaveCounts = new int[nSMCTreeNodes][nGTrees];
        stored_coalIntensities = new double[nSMCTreeNodes][nGTrees];
        */


        // normalize the inverse gamma component weights to sum to 1
        double totalweight = 0.0;
        for (int i = 0; i < priorComponentsInput.get().size(); i++) {
            totalweight += priorComponentsInput.get().get(i).getWeight();
        }
        for (int i = 0; i < priorComponentsInput.get().size(); i++) {
            priorComponentsInput.get().get(i).normalizeWeight(totalweight);
        }
    }



    @Override
    public double calculateLogP() {
        if (debugFlag) {
            numberofdebugchecks++;


        }

        logP = logCoalescenceMigrationProb(getInverseGammaMixture(),
                popPriorScaleInput.get().getValue(), migrationIntensityInput.get().getValue());
        return logP;
    }



    @Override
    protected boolean requiresRecalculation() {
        return true;
    }


    @Override
    public List<String> getArguments() {
        return null;
    }


    @Override
    public List<String> getConditions() {
        return null;
    }


    @Override
    public void sample(final State state, final Random random) {
    }


    /*
    @Override
    public void store() {

        for (int n = 0; n < nSMCTreeNodes; n++) {
            System.arraycopy(coalCounts[n], 0, stored_coalCounts[n], 0, stored_coalCounts[n].length);
            System.arraycopy(migEventsCounts[n], 0, stored_migEventsCounts[n], 0, stored_migEventsCounts[n].length);
            System.arraycopy(migLeaveCounts[n], 0, stored_migLeaveCounts[n], 0, stored_migLeaveCounts[n].length);
            System.arraycopy(coalIntensities[n], 0, stored_coalIntensities[n], 0, stored_coalIntensities[n].length);
        }
        super.store();
    }



    @Override
    public void restore() {

        int [][] tmpCCs = coalCounts;
        coalCounts = stored_coalCounts;
        stored_coalCounts = tmpCCs;

        int [][] tmpECs = migEventsCounts;
        migEventsCounts = stored_migEventsCounts;
        stored_migEventsCounts = tmpECs;

        int [][] tmpLCs = migLeaveCounts;
        migLeaveCounts = stored_migLeaveCounts;
        stored_migLeaveCounts = tmpLCs;

        double [][] tmpCMIs = coalIntensities;
        coalIntensities = stored_coalIntensities;
        stored_coalIntensities = tmpCMIs;
        super.restore();
    }
*/

    /* ************************************************************************************************* */
    /* ************************************************************************************************* */
    /* ************************************************************************************************* */


    // used here and by PopSampleStatistic
    InverseGammaMixture getInverseGammaMixture() {
        int n = priorComponentsInput.get().size();
        InverseGammaMixture igm = new InverseGammaMixture(n);
        for (int c = 0; c < n; c++) {
            igm.setWeight(c, priorComponentsInput.get().get(c).getWeight());
            igm.setAlpha(c, priorComponentsInput.get().get(c).getAlpha());
            igm.setBeta(c, priorComponentsInput.get().get(c).getBeta());
        }
        return igm;
    }



    String getGtreeID(int j) {
        return gTreeWEsInput.get().get(j).getID();
    }

    double getMigrationCount(int j) {
        coalsMigs.updateMigrationsHeightsForOneGTree(j);
        int nSNodes = sTree.getNodeCount();
        int nMigs = 0;
        for (int n = 0; n < nSNodes; n++) {
            ArrayList<NLinChangeInSTreeBranch> linCs = coalsMigs.getMigrationsFromSNodeNrGTree(n, j);
            for (NLinChangeInSTreeBranch linC : linCs) {
                if (linC.getChange() == -1) {
                    nMigs ++;
                }
            }

        }
        return nMigs;
    }

    /*-------------------------------------------------------------------------------------------------*/
    /*------------------------------           private              -----------------------------------*/
    /*-------------------------------------------------------------------------------------------------*/




    private static final Comparator<NLinChangeInSTreeBranch> HEIGHT_ORDER =
            new Comparator<NLinChangeInSTreeBranch>() {
                public int compare(NLinChangeInSTreeBranch nlc1, NLinChangeInSTreeBranch nlc2) {
                    return Double.compare(nlc1.getHeight(), nlc2.getHeight());
                }
            };



    private double logCoalescenceMigrationProb(InverseGammaMixture igm,
                                               double popPriorScale, double migIntensity) {
        double logEP = 0.0;
        for (int j = 0; j < nGTrees; j++) {
             logEP += coalsMigs.updateMigrationsHeightsForOneGTree(j);
        }
        updateCountsNLinsCoalIntensities();
        double logCP = logCoalescentProb(igm, popPriorScale);
        double logMP = logMigrationProb(migIntensity);
        double logP = logCP + logMP + logEP;
        assert Double.isFinite(logP);
        return logP;
    }




    private void updateCountsNLinsCoalIntensities() {
        for (int j = 0; j < nGTrees; j++) {
            // counts
            double coalFactor = gTreesWE.get(j).getCoalFactor();
            for (int n = 0; n < nSMCTreeNodes; n++) {
                coalCounts[n][j] = coalsMigs.getCoalescencesFromSNodeNrGTree(n, j).size();
                ArrayList<NLinChangeInSTreeBranch> nlcs = coalsMigs.getMigrationsFromSNodeNrGTree(n, j);
                int leaveCount = 0;
                for (NLinChangeInSTreeBranch nlc : nlcs) {
                    if (nlc.getChange() < 0) { leaveCount += 1; }
                }
                migEventsCounts[n][j] = nlcs.size();
                migLeaveCounts[n][j] = leaveCount;
            }
            // nlineages
            int [] nLins = new int[nSMCTreeNodes];
            for (int n = 0; n < nSMCTreeTips; n++) {
                nLins[n] = bindings.nLineagesForBeastTipNrAndGtree(n, j);
            }
            updateSubtreeNLineages(nLins, sTree.getRoot(), j);
            // intensities
            for (int n = 0; n < nSMCTreeNodes; n++) {
                ArrayList<NLinChangeInSTreeBranch> migrations = coalsMigs.getMigrationsFromSNodeNrGTree(n, j);
                ArrayList<Double> coalescences = coalsMigs.getCoalescencesFromSNodeNrGTree(n, j);
                ArrayList<NLinChangeInSTreeBranch> nLinChanges = new ArrayList<>();
                double nodeHeight = sTree.getNode(n).getHeight();
                nLinChanges.add(new NLinChangeInSTreeBranch(nodeHeight, 0));
                for (Double coal : coalescences) {
                    nLinChanges.add(new NLinChangeInSTreeBranch(coal, -1));
                }
                for (NLinChangeInSTreeBranch mig : migrations) {
                    assert !sTree.getNode(n).isRoot(); // no migs in/out of root
                    nLinChanges.add(mig);
                }
                Node anc = sTree.getNode(n).getParent();

                if (anc == null) {
                    double gTreeRootHeight = gTrees.get(j).getRoot().getHeight();
                    if (gTreeRootHeight > nodeHeight) {
                        nLinChanges.add(new NLinChangeInSTreeBranch(gTreeRootHeight, 0));
                    } else {
                        assert nLinChanges.size() == 1;
                    }
                } else {
                    nLinChanges.add(new NLinChangeInSTreeBranch(anc.getHeight(), 0));
                }
                Collections.sort(nLinChanges, HEIGHT_ORDER);
                double coalIntensity = 0.0; // gamma_bj
                int n_jbi = nLins[n];
                for (int i = 1; i < nLinChanges.size(); i++) {
                    double c_jbi = nLinChanges.get(i).getHeight() - nLinChanges.get(i-1).getHeight();
                    coalIntensity  +=  c_jbi  *  ( n_jbi * (n_jbi - 1) * 0.5 / coalFactor );
                    n_jbi += nLinChanges.get(i).getChange();
                }
                assert coalIntensity >= 0.0;
                coalIntensities[n][j] = coalIntensity;
            }
        }
    }


    private void debugOutputTreesEmbedding() {
        System.err.flush();
        System.out.println("debugOutputTreesEmbedding()" + " numberofdebugchecks " + numberofdebugchecks);
        System.out.flush();
        System.out.println(Misc.allTreesAsText(sTree, gTrees));
        System.out.flush();
        System.out.println(Misc.embeddingParamsAsText(gTreesWE));
        System.out.flush();
    }



    private void debugOutputGtoSmapping() {
        System.err.flush();
        System.out.println("debugOutputGtoSmapping()" + " numberofdebugchecks " + numberofdebugchecks);
        System.out.println(Misc.gToSmappingAsText(coalsMigs.getGTreeNodeNrToSTreeNodeNr()));
        System.out.flush();
    }


    private void debugOutputCountsIntensities() {
        System.err.flush();
        System.out.println("debugOutputCountsIntensities()" + " numberofdebugchecks " + numberofdebugchecks);
        System.out.println(Misc.CountsIntensitiesAsText(coalCounts, migEventsCounts, migLeaveCounts, coalIntensities));
        System.out.flush();
    }



    private void updateSubtreeNLineages(int [] nlineages, Node node, int j) {
        int lftNLin;
        Node lftNode = node.getLeft();
        if (lftNode.isLeaf()) {
            lftNLin = nlineages[lftNode.getNr()];
        } else {
            updateSubtreeNLineages(nlineages, lftNode, j);
            lftNLin = nlineages[lftNode.getNr()];
        }
        int lftNr = lftNode.getNr();
        lftNLin -= coalCounts[lftNr][j];
        lftNLin += (migEventsCounts[lftNr][j] - 2 * migLeaveCounts[lftNr][j]);
        assert lftNLin >= 0;

        int rgtNLin;
        Node rgtNode = node.getRight();
        if (rgtNode.isLeaf()) {
            rgtNLin = nlineages[rgtNode.getNr()];
        } else {
            updateSubtreeNLineages(nlineages, rgtNode, j);
            rgtNLin = nlineages[rgtNode.getNr()];
        }
        int rgtNr = rgtNode.getNr();
        rgtNLin -= coalCounts[rgtNr][j];
        rgtNLin += (migEventsCounts[rgtNr][j] - 2 * migLeaveCounts[rgtNr][j]);
        assert rgtNLin >= 0;

        nlineages[node.getNr()] = lftNLin + rgtNLin;
    }



    // Uses the counts and intensities to calculate the coalescent-migration log-probability
    // over all SMC-tree branches.
    private double logCoalescentProb(InverseGammaMixture igm, double popPriorScale) {
        double logP = 0.0;

        double [] logCFs = new double[nGTrees];
        for (int j =0; j < nGTrees; j++) {
            logCFs[j] = Math.log(gTreesWE.get(j).getCoalFactor());
        }
        double [] lambdas = igm.getWeights();
        double [] alphas = igm.getAlphas();
        double [] betas = igm.getBetas();
        double [] logProbCpts = new double[lambdas.length];

        for (int n = 0; n < nSMCTreeNodes; n++) {
            int q_b = 0;
            double gamma_b = 0.0;
            double minusLog_r_b = 0.0;
            for (int j= 0; j < nGTrees; j++) {
                q_b += coalCounts[n][j];
                minusLog_r_b += logCFs[j] * coalCounts[n][j];
                gamma_b += coalIntensities[n][j];
            }
            for (int c = 0; c < lambdas.length; c++) {
                logProbCpts[c] = 0.0;
                double sigmabeta_c = popPriorScale * betas[c];
                logProbCpts[c] += Math.log(lambdas[c]);
                logProbCpts[c] += alphas[c] * Math.log(sigmabeta_c);
                logProbCpts[c] -= (alphas[c] + q_b) * Math.log(sigmabeta_c + gamma_b);
                logProbCpts[c] += lnGammaRatiosTable[c][q_b];
            }
            double logP_b = logSumExp(logProbCpts);
            logP_b -= minusLog_r_b;
            logP += logP_b;
        }
        assert Double.isFinite(logP);
        return logP;
    }



    double logMigrationProb(double migIntensity) {
        double logMP = 0.0;
        for (int j = 0; j < migLeaveCounts[0].length; j++) {
            int nMigs_j = 0;
            for (int n = 0; n < migLeaveCounts.length; n++) {
                nMigs_j += migLeaveCounts[n][j];
            }
            if (nMigs_j > 0) {
               // System.out.print("DEBUGGING\n");
            }
            double intensity =  migIntensity * gTrees.get(j).getInternalNodeCount();
            logMP -= intensity;
            logMP += nMigs_j * Math.log(intensity);
            for (int i = 2; i <= nMigs_j; i++) {
                logMP -= Math.log(i);
            }
        }
        assert Double.isFinite(logMP);
        return logMP;
    }


    // returns log(Gamma(a+n)/Gamma(a))
    private double lnRatioGammas(double a, int n) {
        double x = 0.0;
        for (int i = 0; i < n; i++) {
            x += Math.log(a + i);
        }
        return x;
    }


    // Returns log(sum_i e^(x_i)) without over/under flow
    private double logSumExp(double x[]) {
        double maxx = Double.NEGATIVE_INFINITY;
        for (double d : x) {
            if (d > maxx) { maxx = d; }
        }
        double sum = 0.0;
        for (double d : x) {
            sum += Math.exp(d-maxx);
        }
        return maxx + Math.log(sum);
    }
















}