package denim.util;


/*
 *  Created by Graham Jones on 27/10/2016.
 */


import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;
import beast.evolution.tree.TreeInterface;
import beast.evolution.tree.TreeUtils;
import denim.GtreeWithEmbedding;

import java.util.*;


public class CoalescencesMigrations {
    private ArrayList<ArrayList<ArrayList<Double>>> coalescences;
    // coalescences.get(n).get(j) is an array of the heights of coalescences
    // from gtree j which are inside stree branch n.

    private ArrayList<ArrayList<ArrayList<NLinChangeInSTreeBranch>>> migrations;
    // migrations.get(n).get(j) is an array of the heights of migrations
    // in gtree j which enter or leave stree branch n. Each height is accompanied by an integer
    // specifying the change, -1 for coalescences and migrations leaving,
    // +1 for migrations entering.

    private Bindings bindings;

    private int currGTj;
    private TreeInterface sTree;
    private List<GtreeWithEmbedding> gTreesWE;


    // I assume it is more efficient to store these once than to pass them down a recursive routine.
    private int [][] lowestLabelsG;
    private int [] lowestLabelsS;
    private int [] gTreeNodeNrToSTreeNodeNr;
    private double logGTreeLength;
    private double logEmbeddingProb;

    /*************************************************************************/


    private class MigratingBranch {
        Node sourceS;
        Node destS;
        double migHeight;
        double heightInterval;

        MigratingBranch(Node sourceS, Node destS, double migHeight, double heightInterval) {
            this.sourceS = sourceS;
            this.destS = destS;
            this.migHeight = migHeight;
            this.heightInterval = heightInterval;
        }
    }

    public CoalescencesMigrations(TreeInterface sTree, List<GtreeWithEmbedding> gTreesWE, Bindings bindings) {
        this.bindings = bindings;
        this.sTree = sTree;
        this.gTreesWE = gTreesWE;
        initCoalescencesMigrationsArray();
    }


    // updates a 3D array of migrations.
    public double updateMigrationsHeightsForOneGTree(int j) {
        currGTj = j;
        logEmbeddingProb = 0.0;
        // clear old migrations
        for (ArrayList<ArrayList<Double>> hostNodeCoals : coalescences) {
            hostNodeCoals.get(j).clear();
        }
        for (ArrayList<ArrayList<NLinChangeInSTreeBranch>> hostNodeMigs : migrations) {
            hostNodeMigs.get(j).clear();
        }
        Tree gTree = gTreesWE.get(j).getTree();

        makeLowestLabels(gTree.getRoot(), lowestLabelsG[j]);
        makeLowestLabels(sTree.getRoot(), lowestLabelsS);
        gTreeNodeNrToSTreeNodeNr = new int[gTree.getNodeCount()];
        logGTreeLength = Math.log( TreeUtils.getTrunkLength(gTree, gTree.getRoot()) );
        makeCoalescencesMigrationsSubtree(gTree.getRoot());
        assert Double.isFinite(logEmbeddingProb);

        return logEmbeddingProb;
    }

    // Once the migrations array is updated, returns a 1D array of the coalescence
    // migrations from gtree j which are inside stree branch n.
    public ArrayList<NLinChangeInSTreeBranch> getMigrationsFromSNodeNrGTree(int n, int j) {
        return migrations.get(n).get(j);
    }

    public ArrayList<Double> getCoalescencesFromSNodeNrGTree(int n, int j) {
        return coalescences.get(n).get(j);
    }


    // for debugging
    public int [] getGTreeNodeNrToSTreeNodeNr() {
        return gTreeNodeNrToSTreeNodeNr;
    }


    /************************
     * private
     *******************************************/


    private void initCoalescencesMigrationsArray() {
        int numSNodes = sTree.getNodeCount();
        int numGTrees = gTreesWE.size();
        migrations = new ArrayList<>(numSNodes);
        coalescences = new ArrayList<>(numSNodes);
        for (int n = 0; n < numSNodes; n++) {
            migrations.add(new ArrayList<>(numGTrees));
            coalescences.add(new ArrayList<>(numGTrees));
            for (int j = 0; j < numGTrees; j++) {
                migrations.get(n).add(new ArrayList<>(0));
                coalescences.get(n).add(new ArrayList<>(0));
            }
        }
        lowestLabelsG = new int[gTreesWE.size()][];
        for (int j = 0; j < gTreesWE.size(); j++) {
            lowestLabelsG[j] = new int[gTreesWE.get(j).getTree().getNodeCount()];
        }
        lowestLabelsS = new int[sTree.getNodeCount()];
    }


    private void makeLowestLabels(Node x, int [] lowestLabels) {
        if (x.isLeaf()) {
            lowestLabels[x.getNr()] = x.getNr();
        } else {
            makeLowestLabels(x.getLeft(), lowestLabels);
            makeLowestLabels(x.getRight(), lowestLabels);
            int lftLL = lowestLabels[x.getLeft().getNr()];
            int rgtLL = lowestLabels[x.getRight().getNr()];
            assert x.getNr() >= 0;
            assert x.getNr() < lowestLabels.length;
            lowestLabels[x.getNr()]  = Math.min(lftLL, rgtLL);
        }
    }


    //
    private Node makeCoalescencesMigrationsSubtree(final Node hereG) {
        // Find species tree branches lftS and rgtS for children lftG and rgtG
        // of hereG, via recursion.
        final Node lftG = hereG.getLeft();
        final Node lftS;
        if (lftG.isLeaf()) {
            int lftSNr = bindings.smcTipNrFromGTreeTipNr(currGTj, lftG.getNr());
            gTreeNodeNrToSTreeNodeNr[lftG.getNr()] = lftSNr;
            lftS = sTree.getNode(lftSNr);
        } else {
            lftS = makeCoalescencesMigrationsSubtree(lftG);
        }
        final Node rgtG = hereG.getRight();
        final Node rgtS;
        if (rgtG.isLeaf()) {
            int rgtSNr = bindings.smcTipNrFromGTreeTipNr(currGTj, rgtG.getNr());
            gTreeNodeNrToSTreeNodeNr[rgtG.getNr()] = rgtSNr;
            rgtS = sTree.getNode(rgtSNr);
        } else {
            rgtS = makeCoalescencesMigrationsSubtree(rgtG);
        }
        // The MRCA of lftS and rgtS is the most recent species tree branch that hereG
        // could be in if it is compatible.
        final Node hereS = Misc.mrca(lftS, rgtS);
        Node hostS; // will be the species tree branch that hereG goes in
        double heightG = hereG.getHeight();
        if (hereS.getHeight() > heightG) {
            // Coalescence hereG is incompatible.
            // Find the source and destination branches in sTree and the migration height
            MigratingBranch migBr = findMigratingBranch(hereG);
            // Add the lineage changes to the species tree branches.
            migrations.get(migBr.sourceS.getNr()).get(currGTj).add(new NLinChangeInSTreeBranch(migBr.migHeight, -1));
            migrations.get(migBr.destS.getNr()).get(currGTj).add(new NLinChangeInSTreeBranch(migBr.migHeight, 1));
            hostS = migBr.destS;
            logEmbeddingProb += Math.log(migBr.heightInterval);
            logEmbeddingProb -= logGTreeLength;
        } else {
            hostS = hereS;
        }
        // hostS is now the right branch or a descendant of it.
        hostS = Misc.ancestorBranchAt(hostS, heightG);
        // Add hostS to the mapping of gen tree nodes to species tree nodes for later
        // use by migration code. Then add the coalescence (hereG,heightG) to hostS.
        gTreeNodeNrToSTreeNodeNr[hereG.getNr()] = hostS.getNr();
        coalescences.get(hostS.getNr()).get(currGTj).add(heightG);
        return hostS;
    }



    private MigratingBranch findMigratingBranch(Node hereG) {
        double heightG = hereG.getHeight();
        GtreeWithEmbedding gTreeWE = gTreesWE.get(currGTj);
        Tree gTree = gTreeWE.getTree();
        int intNNr = Misc.findInternalNodeIndex(gTree, hereG.getNr());

        // Use migPoint to choose which child of hereG migrates to hereG, and when
        Node sourceG;
        double migPoint = gTreeWE.getMigPoint(intNNr);
        Node children [] = Misc.orderedChildren(hereG, lowestLabelsG[currGTj]);
        double chBrLen0 = heightG - children[0].getHeight();
        double chBrLen1 = heightG - children[1].getHeight();
        double thresh = chBrLen0 / (chBrLen0 + chBrLen1);
        double migHeight;
        if (migPoint < thresh) {
            sourceG = children[0];
            migPoint /= thresh;

        } else {
            sourceG = children[1];
            migPoint -= thresh;
            migPoint /= (1.0-thresh);
        }
        assert migPoint >= 0 && migPoint <= 1.0;
        migHeight =  sourceG.getHeight() + migPoint * (heightG - sourceG.getHeight());
        // Locate the source branch in stree
        Node sourceS = sTree.getNode(gTreeNodeNrToSTreeNodeNr[sourceG.getNr()]);
        sourceS = Misc.ancestorBranchAt(sourceS, migHeight);

        // Locate the destination branch in stree using destinationChoices
        Node destG = Misc.siblingOf(sourceG);
        assert destG != null;
        Node destS = sTree.getNode(gTreeNodeNrToSTreeNodeNr[destG.getNr()]);
        // if destS branch is too recent, find an ancestor
        while (!destS.isRoot()  &&  migHeight > destS.getParent().getHeight()) {
            destS = destS.getParent();
        }
        // if if destS branch is too ancient, pick a descendant
        if (migHeight < destS.getHeight()) {
            ArrayList<Integer> destNNrs = new ArrayList<>();
            collectDestSBranches(destS, migHeight, destNNrs);
            assert destNNrs.size() > 0;
            double dCh = gTreeWE.getDestinationChoice(intNNr);
            dCh *= destNNrs.size();
            for (int idxS = 0; idxS < destNNrs.size(); idxS++) {
                if (dCh >= idxS  && dCh <= idxS + 1) {
                    destS = sTree.getNode(destNNrs.get(idxS));
                }
            }
        }
        assert destS != null;
        assert destS.getHeight() <= migHeight;
        assert destS.getParent().getHeight() >= migHeight;
        assert sourceS.getNr() != destS.getNr();
        return new MigratingBranch(sourceS, destS, migHeight, chBrLen0+chBrLen1);
    }



    // Collect in order given by lowestLabelsS[].
    private void collectDestSBranches(Node destS, double migHeight, ArrayList<Integer> destSNNrs) {
        if (destS.getHeight() <= migHeight) {
            destSNNrs.add(destS.getNr());
        } else {
            Node [] children = Misc.orderedChildren(destS, lowestLabelsS);
            collectDestSBranches(children[0], migHeight, destSNNrs);
            collectDestSBranches(children[1], migHeight, destSNNrs);
        }
    }






}
